"""Exceptions used within aioldata."""


class LDATAException(Exception):
    """Base exception for Leviton Load Data."""


class AuthError(LDATAException):
    """Base exception for authentication."""


class ExceededMaximumRetries(LDATAException):
    """Data unable to be sent after hitting retry limit."""


class InvalidAuth(AuthError):
    """Invalid auth data provided to login."""


class InvalidResidences(LDATAException):
    """No residence data available."""


class InvalidResponse(AuthError):
    """Unexpected response type from API."""


class UnknownAuthError(AuthError):
    """Error occurred when performing auth."""



"""Gathers data from Leviton API."""

__all__ = ["LDATAConnector", "Residence"]

import asyncio
from collections.abc import Generator
from contextlib import asynccontextmanager, contextmanager
import logging
from typing import NamedTuple

import aiohttp
from aiohttp import web_exceptions
from securelogging import LogRedactorMessage

from .auth import InvalidAuth, LDATAAuth
from .const import DEFAULT_HEADERS, MAX_RETRIES
from .errors import ExceededMaximumRetries, InvalidResidences
from .parser import PanelData, parse_data


class Residence(NamedTuple):
    """Data related to a single residence."""

    access: str
    status: str
    person_id: str
    residence_id: int
    residental_account_id: int
    id: int


@contextmanager
def passthrough():
    """Do nothing."""
    yield


class LDATAConnector:
    """Handles communication with Leviton API."""

    def __init__(
        self,
        username: str,
        password: str,
        session: aiohttp.ClientSession | None = None,
        hide_secrets: bool = True,
        three_phase: bool = False,
    ):
        """Initialize the data connector."""
        self.logger = logging.getLogger(f"{__package__}[{username}]")
        self._auth = LDATAAuth(username, password, hide_secrets=hide_secrets)
        self._three_phase = three_phase
        self._web_session = session
        self._close_session = session is None
        self._residences: list[Residence] = []
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        if not session:
            connector = aiohttp.TCPConnector(
                limit_per_host=3,
            )
            self._web_session = aiohttp.ClientSession(connector=connector)

    async def get_headers(self):
        """Get default headers for an API call."""
        headers: dict[str, str] = {**DEFAULT_HEADERS}
        headers["authorization"] = await self._auth.token(self._web_session)
        return headers

    async def _get_residence_information(self) -> list[Residence]:
        """Query Leviton API for all available residences."""
        acct_id = await self._auth.userid(self._web_session)
        url = f"https://my.leviton.com/api/Person/{acct_id}/residentialPermissions"
        res = await self.request(
            "GET",
            url,
        )
        res.raise_for_status()
        data = await res.json()
        if len(data) == 0:
            raise InvalidAuth("No accounts present")
        residences = []
        for elem in data:
            if "residentialAccountId" not in elem:
                continue
            residences.append(
                Residence(
                    access=elem["access"],
                    status=elem["status"],
                    person_id=elem["personId"],
                    residence_id=elem["residenceId"],
                    residental_account_id=elem["residentialAccountId"],
                    id=elem["id"],
                )
            )
        if len(residences) == 0:
            self.logger.warning("Unable to gather residence information")
            raise InvalidResidences
        if residences[0].residence_id is None:
            residence_id = await self.get_primary_residence(
                residences[0].residental_account_id
            )
            residences[0] = Residence(
                access=residences[0].access,
                status=residences[0].status,
                person_id=residences[0].person_id,
                residence_id=residence_id,
                residental_account_id=residences[0].residental_account_id,
                id=residences[0].id,
            )
        self._residences = residences
        return self._residences

    async def get_primary_residence(self, res_account_id: int) -> int:
        """Lookup the primary residence ID."""
        url = f"https://my.leviton.com/api/ResidentialAccounts/{res_account_id}"
        res = await self.request(
            "GET",
            url,
        )
        res.raise_for_status()
        return (await res.json())["primaryResidenceId"]

    async def residental_account_id(self) -> int:
        """Get the account ID for the given credentials."""
        if len(self._residences) == 0:
            await self._get_residence_information()
        return self._residences[0].residentialAccountId

    async def residences(self) -> list[int]:
        """Get a list of residences' IDs for the given credentials."""
        if len(self._residences) == 0:
            await self._get_residence_information()
        return self._residences

    @asynccontextmanager
    async def create_request(
        self, method: str, url: str, **kwargs
    ) -> Generator[aiohttp.ClientResponse, None, None]:
        """Make a request to any path with V2 request method (auth in header).

        Returns a generator with aiohttp ClientResponse.
        """
        headers = await self.get_headers()
        headers.update(kwargs.get("headers", {}))
        kwargs["headers"] = headers
        kwargs["ssl"] = True
        async with self._web_session.request(method, url, **kwargs) as res:
            yield res

    async def request(self, method: str, url: str, **kwargs) -> aiohttp.ClientResponse:
        """Make request on the api and return response data."""
        retries = 0
        with self.secret_logger():
            self.logger.info("Making request [%s] to %s with %s", method, url, kwargs)
        while retries < MAX_RETRIES:
            retries += 1
            if retries > 1:
                retry_wait = 0.25 * retries
                await asyncio.sleep(retry_wait)
            async with self.create_request(method, url, **kwargs) as resp:
                # 503 means the service is temporarily unavailable, back off a bit.
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status in [429, 503]:
                    continue
                # 403 is bad auth
                if resp.status == 403:
                    raise web_exceptions.HTTPForbidden
                await resp.read()
                return resp
        raise ExceededMaximumRetries("Exceeded maximum number of retries")

    async def status(self) -> PanelData:
        """Get the current information for the given panels."""
        api_data = []
        api_data.extend(await self.get_ldata_panels())
        api_data.extend(await self.get_whems_panels())
        return parse_data(api_data, self._three_phase)

    async def get_ldata_panels(self) -> list[dict]:
        """Get all LDATA Panels."""
        all_panels = []
        headers = {"filter": '{"include":["residentialBreakers"]}'}
        for residence in await self.residences():
            url = f"https://my.leviton.com/api/Residences/{residence.residence_id}/residentialBreakerPanels"
            res = await self.request("GET", url, headers=headers)
            for panel in await res.json():
                panel["ModuleType"] = "LDATA"
                all_panels.append(panel)
                self.logger.debug("Panels for %s: %s", residence.residence_id, panel)
        return all_panels

    async def get_whems_panels(self) -> list[dict]:
        """Get all WHEMS IOT Panels."""
        all_panels: list = []
        for residence in await self.residences():
            url = f"https://my.leviton.com/api/Residences/{residence.residence_id}/iotWhems"
            res = await self.request("GET", url)
            for panel in await res.json():
                panel["ModuleType"] = "WHEMS"
                # Make the data look like an LDATA module
                panel["rmsVoltage"] = panel["rmsVoltageA"]
                panel["rmsVoltage2"] = panel["rmsVoltageB"]
                panel["updateVersion"] = panel["version"]
                panel["residentialBreakers"] = await self.get_whems_breakers(
                    panel["id"]
                )
                panel["CTs"] = await self.get_whems_ct(panel["id"])
                all_panels.append(panel)
            self.logger.debug("Panels for %s: %s", residence.residence_id, all_panels)
        return all_panels

    async def get_whems_breakers(self, panel_id: str) -> list[dict]:
        """Get all WHEMS IOT Breakers for the given panel."""
        headers = {
            "filter": "{}",
        }
        url = f"https://my.leviton.com/api/IotWhems/{panel_id}/residentialBreakers"
        res = await self.request("GET", url, headers=headers)
        return await res.json()

    async def get_whems_ct(self, panel_id: str) -> list[dict]:
        """Get all WHEMS IOT CTs for the given panel."""
        headers = {
            "filter": "{}",
        }
        url = f"https://my.leviton.com/api/IotWhems/{panel_id}/iotCts"
        res = await self.request("GET", url, headers=headers)
        return await res.json()

    async def _set_breaker_power(
        self, breaker_id: str, enabled: bool
    ) -> aiohttp.ClientResponse:
        """Set the power state for a given breaker."""
        headers = {
            "referer": f"https://my.leviton.com/home/residential-breakers/{breaker_id}/settings"
        }
        return self.request(
            "POST",
            f"https://my.leviton.com/api/ResidentialBreakers/{breaker_id}",
            json={"remoteOn": enabled},
            headers=headers,
        )

    async def _set_breaker_trip(
        self, breaker_id: str, enabled: bool
    ) -> aiohttp.ClientResponse:
        """Set the trip state for a given breaker."""
        headers = {
            "referer": f"https://my.leviton.com/home/residential-breakers/{breaker_id}/settings"
        }
        return self.request(
            "POST",
            f"https://my.leviton.com/api/ResidentialBreakers/{breaker_id}",
            json={"remoteTrip": enabled},
            headers=headers,
        )

    async def remote_on(self, breaker_id: str) -> aiohttp.ClientResponse:
        """Turn on the given breaker."""
        return self._set_breaker_power(breaker_id, True)

    async def remote_off(self, breaker_id: str) -> aiohttp.ClientResponse:
        """Turn off the given breaker."""
        return self._set_breaker_power(breaker_id, False)

    async def remote_trip(self, breaker_id: str) -> aiohttp.ClientResponse:
        """Trip the given breaker."""
        return self._set_breaker_trip(breaker_id, True)

    async def remote_reset(self, breaker_id: str) -> aiohttp.ClientResponse:
        """Reset (untrip) the given breaker."""
        return self._set_breaker_trip(breaker_id, False)



"""Package for communicating with Leviton Load Data API."""

__all__ = [
    "LEG1_POSITIONS",
    "LEG2_POSITIONS",
    "RMS_VOLTAGE_FACTOR",
    "AuthError",
    "BreakerData",
    "CTData",
    "InvalidAuth",
    "LDATAConnector",
    "LDATAException",
    "PanelData",
    "ParsedData",
    "Residence",
]

from importlib.metadata import PackageNotFoundError, version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


from .connector import LDATAConnector, Residence
from .const import LEG1_POSITIONS, LEG2_POSITIONS, RMS_VOLTAGE_FACTOR
from .errors import AuthError, InvalidAuth, LDATAException
from .parser import BreakerData, CTData, PanelData, ParsedData



import datetime
import asyncio
import logging
import aiohttp
import securelogging
from aioresponses import aioresponses
import pytest
from aioldata import auth
from contextlib import asynccontextmanager

@pytest.fixture
def configured_auth():
    a = auth.LDATAAuth("user", "pass")
    a.auth_data = auth.AuthData(
        token="token",
        userid="userid",
        expires=datetime.datetime.utcnow() + datetime.timedelta(hours=1),
    )
    securelogging.add_secret("token")
    securelogging.add_secret("userid")
    return a


@pytest.fixture(autouse=True)
def reset_secrets():
    securelogging._called_from_test = True
    yield
    securelogging.reset_secrets()


class MockResponse:
    def __init__(self, data, status, has_bad_data=False):
        self._data = data
        self.has_bad_data = has_bad_data
        self.status = status

    async def json(self):
        if not self.has_bad_data:
            return self._data
        else:
            raise ValueError("bad json")

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def __aenter__(self):
        return self


@pytest.mark.parametrize("hide_secrets", [True, False])
def test_LDATAAuth_ensure_no_password(hide_secrets, caplog):
    conf_auth = auth.LDATAAuth("user", "pass", hide_secrets=hide_secrets)
    with conf_auth.secret_logger():
        conf_auth.logger.warning("This is a cool pass")
    if hide_secrets:
        assert "pass" not in caplog.text
    else:
        assert "pass" in caplog.text

@pytest.mark.parametrize(
    ("auth_data", "expected"), [
        (
            None, True
        ),
        (
            auth.AuthData(
                token="token",
                userid="userid",
                expires=datetime.datetime.now() + datetime.timedelta(hours=1),
            ),
            False,
        ),
        (
            auth.AuthData(
                token="token",
                userid="userid",
                expires=datetime.datetime.now() - datetime.timedelta(hours=1),
            ),
            True,
        )
    ]
)
def test_is_expired(auth_data, expected, configured_auth):
    configured_auth.auth_data = auth_data
    assert configured_auth.is_expired == expected


@pytest.mark.asyncio
async def test_token():
    pass


@pytest.mark.asyncio
async def test_userid():
    pass


@pytest.mark.asyncio
async def test_login():
    pass


valid_login_response = MockResponse(
                {
                    "created": "2025-08-08T14:00:00.0000Z",
                    "ttl": 600,
                    "id": "12345",
                    "userId": "67890"
                },
                200
            )

valid_login_expected = auth.AuthData(
                token="12345",
                userid="67890",
                expires=datetime.datetime(2025, 8, 8, 14, 10),
            )

@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("response", "expected", "error", "messages"), [
        # Happy path
        (
            valid_login_response,
            valid_login_expected,
            None,
            ["Successfully authenticated to the API."],
        ),
        # Bad credentials
        (
            MockResponse([], 401), None, auth.InvalidAuth, ["Invalid credentials provided"]
        ),
        # Weird error code
        (
            MockResponse([], 201), None, auth.UnknownAuthError, ["Unable to authenticate. Status Code: 201"]
        ),
        # Bad JSON returned
        (
            MockResponse([], 200, has_bad_data=True), None, auth.InvalidResponse, ["Unexpected data returned during token refresh"]
        ),
    ]
)
async def test_process_login_response(response, expected, error, messages, caplog, configured_auth):
    caplog.set_level(logging.DEBUG)
    if not error:
        assert await configured_auth.process_login_response(response) == expected
    else:
        with pytest.raises(error):
            await configured_auth.process_login_response(response)
    if messages:
        for message in messages:
            assert message in caplog.text



"""Handles parsing of json data."""

__all__ = ["BreakerData", "CTData", "PanelData", "ParsedData", "parse_data"]
from typing import TypedDict

from .const import LEG1_POSITIONS, RMS_VOLTAGE_FACTOR


class BreakerData(TypedDict):
    """Parsed data related to breakers."""

    id: str  # ID for the given breaker
    panel_id: str  # Panel where the breaker exists
    rating: int
    position: int
    name: str
    state: bool
    model: str
    poles: int
    serialNumber: str
    hardware: str
    firmware: str
    canRemoteOn: bool
    remoteState: str
    power: float
    voltage: float
    frequency: float
    current: float
    leg: int
    power1: float
    voltage1: float
    current1: float
    frequency1: float
    power2: float
    voltage2: float
    current2: float
    frequency2: float


CTData = TypedDict(
    "CTData",
    {
        "id": str,
        "name": str,
        "panel_id": str,
        "channel": str,
        "power": float,
        "consumption": float,
        "import": float,
        "current": float,
        "current1": float,
        "current2": float,
    },
)


class PanelData(TypedDict):
    """Parsed data related to Panels."""

    id: str
    name: str
    model: str
    firmware: str
    serialNumber: str
    voltage: float
    voltage1: float
    frequency1: float
    frequency2: float


class ParsedData(TypedDict):
    """Parsed API data.

    This data includes a <panel_id>_totalPower key for each panel.
    """

    breakers: dict[str, BreakerData]
    cts: dict[str, CTData]
    panels: list[PanelData]


def none_to_zero(data, key) -> float:
    """Convert a value to a float and replace None with 0.0."""
    try:
        result = float(data.get(key, None))
    except (TypeError, ValueError):
        result = 0.0
    return result


def parse_data(api_data: list[dict], three_phase: bool = False) -> ParsedData:
    """Parse the API data into a well-formatted object."""
    breakers: dict[str, BreakerData] = {}
    cts: dict[str, CTData] = {}
    panels: list[PanelData] = []
    panel_power: dict[str, float] = {}
    for panel in api_data:
        panel_data = PanelData()
        panel_data["firmware"] = panel["updateVersion"]
        panel_data["model"] = panel["model"]
        panel_data["id"] = panel["id"]
        panel_data["name"] = panel["name"]
        panel_data["serialNumber"] = panel["id"]
        if not three_phase:
            panel_data["voltage"] = (
                float(panel["rmsVoltage"]) + float(panel["rmsVoltage2"])
            ) / 2.0
        else:
            panel_data["voltage"] = (
                float(panel["rmsVoltage"]) * RMS_VOLTAGE_FACTOR
            ) + (float(panel["rmsVoltage2"]) * RMS_VOLTAGE_FACTOR)
        panel_data["voltage1"] = float(panel["rmsVoltage"])
        panel_data["voltage2"] = float(panel["rmsVoltage2"])
        panel_data["frequency1"] = float(none_to_zero(panel, "frequencyA"))
        panel_data["frequency2"] = float(none_to_zero(panel, "frequencyB"))
        if panel_data["frequency1"] == 0:
            panel_data["frequency1"] = 0
            panel_data["frequency2"] = 0
            for breaker in panel["residentialBreakers"]:
                if breaker["position"] in LEG1_POSITIONS:
                    if float(none_to_zero(breaker, "lineFrequency")) > 0:
                        panel_data["frequency1"] = float(
                            none_to_zero(breaker, "lineFrequency")
                        )
                    if float(none_to_zero(breaker, "lineFrequency2")) > 0:
                        panel_data["frequency2"] = float(
                            none_to_zero(breaker, "lineFrequency2")
                        )
                else:
                    if float(none_to_zero(breaker, "lineFrequency")) > 0:
                        panel_data["frequency2"] = float(
                            none_to_zero(breaker, "lineFrequency")
                        )
                    if float(none_to_zero(breaker, "lineFrequency2")) > 0:
                        panel_data["frequency1"] = float(
                            none_to_zero(breaker, "lineFrequency2")
                        )
                if panel_data["frequency1"] != 0 and panel_data["frequency2"] != 0:
                    break
        if panel_data["frequency2"] == 0:
            panel_data["frequency2"] = panel_data["frequency1"]
        panel_data["frequency"] = (
            float(panel_data["frequency1"]) + float(panel_data["frequency2"])
        ) / 2
        panels.append(panel_data)
        if "CTs" in panel:
            cts.update(parse_cts(panel["CTs"], panel["id"]))
        parsed_breaker_data = parse_breakers(
            panel["residentialBreakers"], panel["id"], three_phase=three_phase
        )
        panel_power[panel["id"] + "totalPower"] = sum(
            breaker["power"]
            for breaker in parsed_breaker_data
            if isinstance(breaker["power"], float)
        )
        breakers.update(parsed_breaker_data)
    full_data = ParsedData(breakers=breakers, cts=cts, panels=panels)
    full_data.update(panel_power)
    return full_data


def parse_breakers(
    api_breakers: list, panel_id: str, three_phase: bool
) -> dict[str, BreakerData]:
    breakers: dict[str, BreakerData] = {}
    for breaker in api_breakers:
        if breaker["model"] in [None, "NONE-1", "NONE-2"]:
            continue
        breaker_data = BreakerData()
        breaker_data["panel_id"] = panel_id
        breaker_data["rating"] = breaker["currentRating"]
        breaker_data["position"] = breaker["position"]
        breaker_data["name"] = breaker["name"]
        breaker_data["state"] = breaker["currentState"]
        breaker_data["id"] = breaker["id"]
        breaker_data["model"] = breaker["model"]
        breaker_data["poles"] = breaker["poles"]
        breaker_data["serialNumber"] = breaker["serialNumber"]
        breaker_data["hardware"] = breaker["hwVersion"]
        breaker_data["firmware"] = breaker["firmwareVersionMeter"]
        breaker_data["canRemoteOn"] = bool(breaker["canRemoteOn"])
        breaker_data["remoteState"] = (
            "RemoteON"
            if breaker["remoteState"] in ["", "RemoteOn"]
            else breaker["remoteState"]
        )
        breaker_data["power"] = none_to_zero(breaker, "power") + none_to_zero(
            breaker, "power2"
        )
        if not three_phase or (breaker["poles"] == 1):
            breaker_data["voltage"] = none_to_zero(
                breaker, "rmsVoltage"
            ) + none_to_zero(breaker, "rmsVoltage2")
        else:
            breaker_data["voltage"] = (
                none_to_zero(breaker, "rmsVoltage") * RMS_VOLTAGE_FACTOR
            ) + (none_to_zero(breaker, "rmsVoltage2") * RMS_VOLTAGE_FACTOR)

        if breaker["poles"] == 2:
            breaker_data["frequency"] = (
                none_to_zero(breaker, "lineFrequency")
                + none_to_zero(breaker, "lineFrequency2")
            ) / 2.0
            breaker_data["current"] = (
                none_to_zero(breaker, "rmsCurrent")
                + none_to_zero(breaker, "rmsCurrent2")
            ) / 2
        else:
            breaker_data["frequency"] = none_to_zero(breaker, "lineFrequency")
            breaker_data["current"] = none_to_zero(
                breaker, "rmsCurrent"
            ) + none_to_zero(breaker, "rmsCurrent2")
        if breaker["position"] in LEG1_POSITIONS:
            breaker_data["leg"] = 1
            breaker_data["power1"] = none_to_zero(breaker, "power")
            breaker_data["power2"] = none_to_zero(breaker, "power2")
            breaker_data["voltage1"] = none_to_zero(breaker, "rmsVoltage")
            breaker_data["voltage2"] = none_to_zero(breaker, "rmsVoltage2")
            breaker_data["current1"] = none_to_zero(breaker, "rmsCurrent")
            breaker_data["current2"] = none_to_zero(breaker, "rmsCurrent2")
            breaker_data["frequency1"] = none_to_zero(breaker, "lineFrequency")
            breaker_data["frequency2"] = none_to_zero(breaker, "lineFrequency2")
        else:
            breaker_data["leg"] = 2
            breaker_data["power1"] = none_to_zero(breaker, "power2")
            breaker_data["power2"] = none_to_zero(breaker, "power")
            breaker_data["voltage1"] = none_to_zero(breaker, "rmsVoltage2")
            breaker_data["voltage2"] = none_to_zero(breaker, "rmsVoltage")
            breaker_data["current1"] = none_to_zero(breaker, "rmsCurrent2")
            breaker_data["current2"] = none_to_zero(breaker, "rmsCurrent")
            breaker_data["frequency1"] = none_to_zero(breaker, "lineFrequency2")
            breaker_data["frequency2"] = none_to_zero(breaker, "lineFrequency")
        breakers[breaker["id"]] = breaker_data
    return breakers


def parse_cts(cts_data: list[dict], panel_id: str) -> dict[str, CTData]:
    parsed_data: dict[str, CTData] = {}
    for ct in cts_data:
        if ct["usageType"] == "NOT_USED":
            continue
        ct_data = CTData()
        ct_data["name"] = ct["usageType"]
        ct_data["id"] = str(ct["id"])
        ct_data["panel_id"] = panel_id
        ct_data["channel"] = str(ct["channel"])
        ct_data["power"] = none_to_zero(ct, "activePower") + none_to_zero(
            ct, "activePower2"
        )
        ct_data["consumption"] = none_to_zero(ct, "energyConsumption") + none_to_zero(
            ct, "energyConsumption2"
        )
        ct_data["import"] = none_to_zero(ct, "energyImport") + none_to_zero(
            ct, "energyImport2"
        )
        ct_data["current"] = (
            none_to_zero(ct, "rmsCurrent") + none_to_zero(ct, "rmsCurrent2")
        ) / 2
        ct_data["current1"] = none_to_zero(ct, "rmsCurrent")
        ct_data["current2"] = none_to_zero(ct, "rmsCurrent2")
        # Add the CT to the list.
        parsed_data[ct_data["id"]] = ct_data
    return parsed_data



"""Handles auth for Leviton Load Center."""

import asyncio
from contextlib import contextmanager
from datetime import datetime, timedelta
import logging
from typing import NamedTuple

from aiohttp import ClientResponse, ClientSession, ContentTypeError
from securelogging import LogRedactorMessage, add_secret, remove_secret

from .const import DEFAULT_HEADERS, MAX_RETRIES
from .errors import (
    ExceededMaximumRetries,
    InvalidAuth,
    InvalidResponse,
    UnknownAuthError,
)


@contextmanager
def passthrough():
    """Do nothing."""
    yield


class AuthData(NamedTuple):
    """Information related to the current Auth."""

    token: str
    userid: str
    expires: datetime


class LDATAAuth:
    """Handles auth-related information to Leviton API."""

    def __init__(self, username, password, hide_secrets: bool = False):
        """Initialize auth data."""
        self.logger = logging.getLogger(f"{__package__}[{username}]")
        self.username = username
        self.password = password
        self.auth_data: AuthData | None = None
        self._async_lock: asyncio.Lock = asyncio.Lock()
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        add_secret(password)
        with self.secret_logger():
            self.logger.info("Initializing connection for %s", username)

    @property
    def is_expired(self):
        """Determine if the current login data is expired."""
        if self.auth_data is None:
            return True
        return self.auth_data.expires <= datetime.now()

    async def token(self, client: ClientSession) -> str:
        """Get the auth token for making requests.

        If the token does not exist, perform the login to generate the token
        """
        if self.is_expired:
            await self.login(client)
        return self.auth_data.token

    async def userid(self, client: ClientSession) -> str:
        """Get the logged-in user for making request."""
        if self.is_expired:
            await self.login(client)
        return self.auth_data.userid

    async def login(self, client: ClientSession) -> AuthData:
        """Perform the login request."""
        async with self._async_lock:
            old_secrets = []
            if self.auth_data is not None:
                old_secrets = [self.auth_data.token, self.auth_data.userid]
            data = {"email": self.username, "password": self.password}
            retries = 0
            success = False
            while retries < MAX_RETRIES and not success:
                try:
                    async with client.post(
                        "https://my.leviton.com/api/Person/login?include=user",
                        headers=DEFAULT_HEADERS,
                        json=data,
                    ) as response:
                        self.auth_data = await self.process_login_response(response)
                        success = True
                except (InvalidResponse, UnknownAuthError):
                    retries += 1
                    await asyncio.sleep(1)
            if not success:
                raise ExceededMaximumRetries
            # Swap out the secrets
            add_secret(self.auth_data.token)
            add_secret(self.auth_data.userid)
            for secret in old_secrets:
                remove_secret(secret)
        return self.auth_data

    async def process_login_response(self, response: ClientResponse) -> AuthData:
        """Extract the pertinent information from the login response."""
        if response.status == 401:
            self.logger.warning("Invalid credentials provided")
            raise InvalidAuth
        if response.status != 200:
            self.logger.warning(
                "Unable to authenticate. Status Code: %s",
                response.status,
            )
            raise UnknownAuthError
        try:
            resp_json = await response.json()
        except (ValueError, ContentTypeError) as err:
            self.logger.exception("Unexpected data returned during token refresh")
            raise InvalidResponse from err
        created_dt = datetime.strptime(resp_json["created"], "%Y-%m-%dT%H:%M:%S.%fZ")
        auth_data = AuthData(
            token=resp_json["id"],
            userid=resp_json["userId"],
            expires=created_dt + timedelta(seconds=resp_json["ttl"]),
        )
        self.logger.info("Successfully authenticated to the API.")
        return auth_data
